/*
Copyright 2020 The Vitess Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package engine

import (
	"vitess.io/vitess/go/sqltypes"
	"vitess.io/vitess/go/vt/key"
	vtrpcpb "vitess.io/vitess/go/vt/proto/vtrpc"
	"vitess.io/vitess/go/vt/vterrors"
	"vitess.io/vitess/go/vt/vtgate/vindexes"

	querypb "vitess.io/vitess/go/vt/proto/query"
)

var _ Primitive = (*Send)(nil)

// Send is an operator to send query to the specific keyspace, tabletType and destination
type Send struct {
	// Keyspace specifies the keyspace to send the query to.
	Keyspace *vindexes.Keyspace

	// TargetDestination specifies an explicit target destination to send the query to.
	TargetDestination key.Destination

	// Query specifies the query to be executed.
	Query string

	// IsDML specifies how to deal with autocommit behaviour
	IsDML bool

	// SingleShardOnly specifies that the query must be send to only single shard
	SingleShardOnly bool

	// ShardNameNeeded specified that the shard name is added to the bind variables
	ShardNameNeeded bool

	noInputs
}

// ShardName as key for setting shard name in bind variables map
const ShardName = "__vt_shard"

//NeedsTransaction implements the Primitive interface
func (s *Send) NeedsTransaction() bool {
	return s.IsDML
}

// RouteType implements Primitive interface
func (s *Send) RouteType() string {
	if s.IsDML {
		return "SendDML"
	}

	return "Send"
}

// GetKeyspaceName implements Primitive interface
func (s *Send) GetKeyspaceName() string {
	return s.Keyspace.Name
}

// GetTableName implements Primitive interface
func (s *Send) GetTableName() string {
	return ""
}

// Execute implements Primitive interface
func (s *Send) Execute(vcursor VCursor, bindVars map[string]*querypb.BindVariable, wantfields bool) (*sqltypes.Result, error) {
	rss, _, err := vcursor.ResolveDestinations(s.Keyspace.Name, nil, []key.Destination{s.TargetDestination})
	if err != nil {
		return nil, err
	}

	if !s.Keyspace.Sharded && len(rss) != 1 {
		return nil, vterrors.Errorf(vtrpcpb.Code_FAILED_PRECONDITION, "Keyspace does not have exactly one shard: %v", rss)
	}

	if s.SingleShardOnly && len(rss) != 1 {
		return nil, vterrors.Errorf(vtrpcpb.Code_INVALID_ARGUMENT, "Unexpected error, DestinationKeyspaceID mapping to multiple shards: %s, got: %v", s.Query, s.TargetDestination)
	}

	queries := make([]*querypb.BoundQuery, len(rss))
	for i, rs := range rss {
		bv := bindVars
		if s.ShardNameNeeded {
			bv = copyBindVars(bindVars)
			bv[ShardName] = sqltypes.StringBindVariable(rs.Target.Shard)
		}
		queries[i] = &querypb.BoundQuery{
			Sql:           s.Query,
			BindVariables: bv,
		}
	}

	canAutocommit := false
	if s.IsDML {
		canAutocommit = len(rss) == 1 && vcursor.AutocommitApproval()
	}

	rollbackOnError := s.IsDML // for non-dml queries, there's no need to do a rollback
	result, errs := vcursor.ExecuteMultiShard(rss, queries, rollbackOnError, canAutocommit)
	err = vterrors.Aggregate(errs)
	if err != nil {
		return nil, err
	}
	return result, nil
}

func copyBindVars(in map[string]*querypb.BindVariable) map[string]*querypb.BindVariable {
	out := make(map[string]*querypb.BindVariable, len(in)+1)
	for k, v := range in {
		out[k] = v
	}
	return out
}

// StreamExecute implements Primitive interface
func (s *Send) StreamExecute(vcursor VCursor, bindVars map[string]*querypb.BindVariable, wantfields bool, callback func(*sqltypes.Result) error) error {
	rss, _, err := vcursor.ResolveDestinations(s.Keyspace.Name, nil, []key.Destination{s.TargetDestination})
	if err != nil {
		return err
	}

	if !s.Keyspace.Sharded && len(rss) != 1 {
		return vterrors.Errorf(vtrpcpb.Code_FAILED_PRECONDITION, "Keyspace does not have exactly one shard: %v", rss)
	}

	if s.SingleShardOnly && len(rss) != 1 {
		return vterrors.Errorf(vtrpcpb.Code_INVALID_ARGUMENT, "Unexpected error, DestinationKeyspaceID mapping to multiple shards: %s, got: %v", s.Query, s.TargetDestination)
	}

	multiBindVars := make([]map[string]*querypb.BindVariable, len(rss))
	for i, rs := range rss {
		bv := bindVars
		if s.ShardNameNeeded {
			bv = copyBindVars(bindVars)
			bv[ShardName] = sqltypes.StringBindVariable(rs.Target.Shard)
		}
		multiBindVars[i] = bv
	}
	errors := vcursor.StreamExecuteMulti(s.Query, rss, multiBindVars, callback)
	return vterrors.Aggregate(errors)
}

// GetFields implements Primitive interface
func (s *Send) GetFields(vcursor VCursor, bindVars map[string]*querypb.BindVariable) (*sqltypes.Result, error) {
	qr, err := s.Execute(vcursor, bindVars, false)
	if err != nil {
		return nil, err
	}
	qr.Rows = nil
	return qr, nil
}

func (s *Send) description() PrimitiveDescription {
	other := map[string]interface{}{
		"Query": s.Query,
		"Table": s.GetTableName(),
	}
	if s.IsDML {
		other["IsDML"] = true
	}
	if s.SingleShardOnly {
		other["SingleShardOnly"] = true
	}
	if s.ShardNameNeeded {
		other["ShardNameNeeded"] = true
	}
	return PrimitiveDescription{
		OperatorType:      "Send",
		Keyspace:          s.Keyspace,
		TargetDestination: s.TargetDestination,
		Other:             other,
	}
}
